import math
from typing import Optional, Tuple

from .data_types import (
    FlexibleSlicesType,
    SliceSlicesType,
    Integer,
    SingleFlexibleSliceType,
)


def _sanitise_slice(start: int, stop: int, step: int) -> Tuple[int, int, int]:
    step_count = math.ceil(max((stop - start) / step, 0))
    stop = start + step_count * step
    if step_count:
        stop += int(math.copysign(1, step)) - step

    return start, stop, step


def sanitise_slice(
    start: Optional[int], stop: Optional[int], step: Optional[int], arr_size: int
) -> Tuple[int, int, int]:
    """
    Convert slices into a sane format
    0 is always before the first number and arr_size is always after the last number.
    """
    # set default values
    if step is None:
        step = 1
    elif step == 0:
        raise Exception("step cannot be 0")

    if start is None:
        if step > 0:
            start = 0
        else:
            start = arr_size - 1
    elif start < 0:
        start = arr_size + start

    if stop is None:
        if step > 0:
            stop = arr_size
        else:
            stop = -1
    elif stop < 0:
        stop = arr_size + stop

    if step < 0:
        start += 1
        stop += 1

    # cap at the ends
    if step > 0:
        start = max(0, start)
        stop = min(stop, arr_size)
    elif step < 0:
        start = min(arr_size, start)
        stop = max(stop, 0)

    return _sanitise_slice(start, stop, step)


def sanitise_unbounded_slice(
    start: Optional[int],
    stop: Optional[int],
    step: Optional[int],
    default_min: int,  # the minimum value to default to
    default_max: int,  # the maximum value to default to
) -> Tuple[int, int, int]:
    if step is None:
        step = 1
    elif step == 0:
        raise Exception("step cannot be 0")

    # cap at the normal limits
    if step > 0:
        if start is None:
            start = default_min
        if stop is None:
            stop = default_max
    else:
        if start is None:
            start = default_max
        if stop is None:
            stop = default_min

    if step < 0:
        start += 1
        stop += 1

    return _sanitise_slice(start, stop, step)


def unsanitise_slice(
    start: int, stop: int, step: int, arr_size: int
) -> Tuple[int, int, int]:
    """Convert sanitised slices back to the normal format."""
    if step < 0:
        start -= 1
        stop -= 1
        if start < 0:
            start -= arr_size
        if stop < 0:
            stop -= arr_size

    return start, stop, step


def stack_sanitised_slices(
    start1: int, stop1: int, step1: int, start2: int, stop2: int, step2: int
) -> Tuple[int, int, int]:
    step3 = step1 * step2
    start3 = start1 + start2 * step1
    stop3 = start1 + stop2 * step1

    return _sanitise_slice(start3, stop3, step3)


def unpack_slice(item: slice):
    return item.start, item.stop, item.step


def to_slice(item: SingleFlexibleSliceType) -> slice:
    if isinstance(item, Integer):
        item = int(item)
        return slice(item, item + 1, 1)
    elif isinstance(item, slice):
        step = item.step
        if step == 0:
            raise Exception("Step of 0 is invalid")
        return slice(item.start, item.stop, 1 if step is None else step)
    else:
        raise Exception(f"Unsupported slice item {item}")


def multi_to_slice(slices: FlexibleSlicesType) -> SliceSlicesType:
    return tuple(to_slice(item) for item in slices)


def get_sliced_array_size(
    start: Optional[int], stop: Optional[int], step: Optional[int], arr_size
) -> int:
    """Find the size of the array the slice would produce from an array of size arr_size"""
    start, stop, step = sanitise_slice(start, stop, step, arr_size)
    return get_sanitised_sliced_array_size(start, stop, step)


def get_sanitised_sliced_array_size(start: int, stop: int, step: int):
    """Find the size of a slice that has been pre-sanitised"""
    return math.ceil((stop - start) / step)



import pkgutil
import sys
import os

import amulet
from amulet import level
from amulet import operations

AMULET_PATH = amulet.__path__[0]

hiddenimports = (
    [name for _, name, _ in pkgutil.walk_packages(level.__path__, level.__name__ + ".")]
    + [
        name
        for _, name, _ in pkgutil.walk_packages(
            operations.__path__, operations.__name__ + "."
        )
    ]
    + ["amulet.api.structure"]
)



from __future__ import annotations

import logging

import numpy
from amulet_nbt import IntArrayTag
from amulet.api.chunk import Chunk
from .base_anvil_interface import ChunkDataType
from .anvil_1934 import Anvil1934Interface as ParentInterface

log = logging.getLogger(__name__)


class Anvil2203Interface(ParentInterface):
    """
    Made biomes 3D
    """

    @staticmethod
    def minor_is_valid(key: int):
        return 2203 <= key < 2529

    def _decode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        biomes = self.get_layer_obj(data, self.Biomes, pop_last=True)
        if isinstance(biomes, IntArrayTag):
            if (len(biomes) / 16) % 4:
                log.error(
                    f"The biome array size must be 4x4x4xN but got an array of size {biomes.np_array.size}"
                )
            else:
                arr = numpy.transpose(
                    biomes.np_array.astype(numpy.uint32).reshape((-1, 4, 4)),
                    (2, 0, 1),
                )  # YZX -> XYZ
                chunk.biomes = {
                    sy + floor_cy: arr
                    for sy, arr in enumerate(
                        numpy.split(
                            arr,
                            arr.shape[1] // 4,
                            1,
                        )
                    )
                }

    def _encode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        if chunk.status.value > -0.7:
            chunk.biomes.convert_to_3d()
            self.set_layer_obj(
                data,
                self.Biomes,
                IntArrayTag(
                    numpy.transpose(
                        numpy.asarray(
                            chunk.biomes[
                                :, floor_cy * 4 : (floor_cy + height_cy) * 4, :
                            ]
                        ).astype(numpy.uint32),
                        (1, 2, 0),
                    ).ravel()  # YZX -> XYZ
                ),
            )


export = Anvil2203Interface



"""
local history manager
    key based storage
        When a piece of data is loaded from disk it needs to be stored as revision 0
        If a piece of data is deleted it should be stored as None to specify that the data has been deleted
        If a piece of data was created rather than loaded from disk the revision 0 should be populated as None
            to specify that the revision 0 did not exist and the data should be added as revision 1
        It needs to store the revision index that is current within the editor.
        It also needs to store the revision that is saved to disk so that we know what has changed compared to
            the current save state rather than the original save state.
        If the above two indexes do not match then the data has been changed compared to the saved version so it needs saving.
    global storage
        We need to store a list of pools of keys that have changed for each undo point.
        We also need to store the index of which pool is current. This will get changed when undoing/redoing to keep track.


    When the creation of an undo point is requested
        we should scan all of the objects and see if the `changed` flag has been set to `True`.
        If this is the case the data has been changed and we should create a backup of it.
            make sure to reset the `changed` flag to False for the cached version
        A new revision for each entry is created and the current revision index is incremented.
        A pool of keys that have changed should be created.
            When undoing this change we look at this pool to see what changed and undo each of those.
            The same applies when redoing
        Finally the `changed` flags for each of the entries should be changed to False so that the next
            undo point only catches what has changed since this undo point.
            The better solution here is to empty the temporary database and re-populate from the latest cache
                revision where the `changed` flag was already set to False. This also solves the issue where
                the data was modified but the `changed` flag was not set to True by the script.

global history manager
    would store each of the local history managers
    would be the entry
"""

from .changeable import Changeable



from __future__ import annotations

import logging
from typing import Dict, Set, Tuple, Iterable, Optional, TYPE_CHECKING

import numpy
from amulet_nbt import (
    CompoundTag,
    ListTag,
    StringTag,
    LongArrayTag,
    LongTag,
    ShortTag,
)

import amulet
from amulet.api.data_types import AnyNDArray, BlockCoordinates
from amulet.api.block import Block
from amulet.api.chunk import StatusFormats
from .base_anvil_interface import ChunkDataType, ChunkPathType
from .anvil_0 import Anvil0Interface as ParentInterface
from amulet.utils.world_utils import (
    decode_long_array,
    encode_long_array,
)

if TYPE_CHECKING:
    from amulet.api.chunk import Chunk

log = logging.getLogger(__name__)


class Anvil1444Interface(ParentInterface):
    """
    Moved TerrainPopulated and LightPopulated to Status
    Made blocks paletted
    Added more tick tags
    Added structures tag
    """

    TerrainPopulated = None
    LightPopulated = None
    Status: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Status", StringTag)],
        StringTag("full"),
    )

    ToBeTicked: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("ToBeTicked", ListTag)],
        ListTag,
    )

    LiquidTicks: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("LiquidTicks", ListTag)],
        ListTag,
    )
    LiquidsToBeTicked: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("LiquidsToBeTicked", ListTag)],
        ListTag,
    )

    PostProcessing: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("PostProcessing", ListTag)],
        ListTag,
    )

    Structures: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Structures", CompoundTag)],
        CompoundTag,
    )

    LongArrayDense = True

    def __init__(self):
        super().__init__()
        self._set_feature("status", StatusFormats.Java_13)

        self._register_decoder(self._decode_fluid_ticks)
        self._register_decoder(self._decode_post_processing)
        self._register_decoder(self._decode_structures)

        self._register_encoder(self._encode_fluid_ticks)
        self._register_encoder(self._encode_post_processing)
        self._register_encoder(self._encode_structures)

    @staticmethod
    def minor_is_valid(key: int):
        return 1444 <= key < 1466

    def _decode_status(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.status = self.get_layer_obj(data, self.Status, pop_last=True).py_str

    def _decode_block_section(
        self, section: CompoundTag
    ) -> Optional[Tuple[numpy.ndarray, list]]:
        if "Palette" not in section:  # 1.14 makes block_palette/blocks optional.
            return None
        section_palette = self._decode_block_palette(section.pop("Palette"))
        decoded = decode_long_array(
            section.get_long_array("BlockStates").np_array,
            4096,
            max(4, (len(section_palette) - 1).bit_length()),
            dense=self.LongArrayDense,
        ).astype(numpy.uint32)
        arr = numpy.transpose(decoded.reshape((16, 16, 16)), (2, 0, 1))
        return arr, section_palette

    def _decode_blocks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        blocks: Dict[int, numpy.ndarray] = {}
        palette = [Block(namespace="minecraft", base_name="air")]

        for cy, section in self._iter_sections(data):
            data = self._decode_block_section(section)
            if data is not None:
                arr, section_palette = data
                blocks[cy] = arr + len(palette)
                palette += section_palette

        np_palette, inverse = numpy.unique(palette, return_inverse=True)
        np_palette: numpy.ndarray
        inverse: numpy.ndarray
        inverse = inverse.astype(numpy.uint32)
        for cy in blocks:
            blocks[cy] = inverse[blocks[cy]]
        chunk.blocks = blocks
        chunk.misc["block_palette"] = np_palette

    @staticmethod
    def _decode_block_palette(palette: ListTag) -> list:
        blockstates = []
        for entry in palette:
            namespace, base_name = entry.get_string("Name").py_str.split(":", 1)
            properties = entry.get_compound("Properties", CompoundTag({})).py_dict
            block = Block(
                namespace=namespace, base_name=base_name, properties=properties
            )
            blockstates.append(block)
        return blockstates

    @staticmethod
    def _decode_to_be_ticked(ticks: ListTag, floor_cy: int) -> Set[BlockCoordinates]:
        section_ticks: ListTag
        pos: ShortTag
        ticks_out = set()
        for cy, section_ticks in enumerate(ticks):
            block_cy = (floor_cy + cy) << 4
            for pos in section_ticks:
                pos_int = pos.py_int
                # TODO: check if these are correct. The order may be wrong.
                x = pos_int & 0xF
                y = (pos_int >> 4) & 0xF
                z = (pos_int >> 8) & 0xF
                ticks_out.add((x, block_cy + y, z))
        return ticks_out

    def _decode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        super()._decode_block_ticks(chunk, data, floor_cy, height_cy)
        chunk.misc["to_be_ticked"] = self._decode_to_be_ticked(
            self.get_layer_obj(data, self.ToBeTicked, pop_last=True), floor_cy
        )

    def _decode_fluid_ticks(self, chunk: Chunk, data: ChunkDataType, floor_cy: int, __):
        chunk.misc.setdefault("fluid_ticks", {}).update(
            self._decode_ticks(
                self.get_layer_obj(data, self.LiquidTicks, pop_last=True)
            )
        )
        chunk.misc["liquids_to_be_ticked"] = self._decode_to_be_ticked(
            self.get_layer_obj(data, self.LiquidsToBeTicked, pop_last=True), floor_cy
        )

    def _decode_post_processing(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["post_processing"] = self._decode_to_be_ticked(
            self.get_layer_obj(data, self.PostProcessing, pop_last=True), floor_cy
        )

    def _decode_structures(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        chunk.misc["structures"] = self.get_layer_obj(
            data, self.Structures, pop_last=True
        )

    def _encode_status(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        # Order the float value based on the order they would be run. Newer replacements for the same come just after
        # to save back find the next lowest valid value.
        status = chunk.status.as_type(self._features["status"])
        self.set_layer_obj(data, self.Status, StringTag(status))

    def _encode_inhabited_time(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data, self.InhabitedTime, LongTag(chunk.misc.get("inhabited_time", 0))
        )

    def _encode_block_section(
        self,
        chunk: Chunk,
        sections: Dict[int, CompoundTag],
        palette: AnyNDArray,
        cy: int,
    ) -> bool:
        block_sub_array = numpy.transpose(
            chunk.blocks.get_sub_chunk(cy), (1, 2, 0)
        ).ravel()

        sub_palette_, block_sub_array = numpy.unique(
            block_sub_array, return_inverse=True
        )
        sub_palette = self._encode_block_palette(palette[sub_palette_])
        if (
            len(sub_palette) == 1
            and sub_palette[0].get_string("Name").py_str == "minecraft:air"
        ):
            return False

        section = sections.setdefault(cy, CompoundTag())
        section["BlockStates"] = LongArrayTag(
            encode_long_array(
                block_sub_array, dense=self.LongArrayDense, min_bits_per_entry=4
            )
        )
        section["Palette"] = sub_palette

    @staticmethod
    def _encode_block_palette(blockstates: Iterable[Block]) -> ListTag:
        palette = ListTag()
        for block in blockstates:
            entry = CompoundTag()
            entry["Name"] = StringTag(f"{block.namespace}:{block.base_name}")
            if block.properties:
                string_properties = {
                    k: v
                    for k, v in block.properties.items()
                    if isinstance(v, StringTag)
                }
                if string_properties:
                    entry["Properties"] = CompoundTag(string_properties)
            palette.append(entry)
        return palette

    @staticmethod
    def _encode_to_be_ticked(
        ticks: Set[BlockCoordinates], floor_cy: int, height_cy: int
    ) -> ListTag:
        ceil_cy = floor_cy + height_cy
        ticks_out = ListTag([ListTag([], 2) for _ in range(floor_cy, ceil_cy)])
        if isinstance(ticks, set):
            for k in ticks:
                try:
                    (x, y, z) = k
                    cy = y >> 4
                    if floor_cy <= cy < ceil_cy:
                        x = x & 15
                        y = y & 15
                        z = z & 15
                        ticks_out[cy].append(ShortTag((z << 8) + (y << 4) + x))
                except Exception:
                    log.error(f"Could not serialise tick data {k}")
        return ticks_out

    def _encode_block_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        super()._encode_block_ticks(chunk, data, floor_cy, height_cy)
        self.set_layer_obj(
            data,
            self.ToBeTicked,
            self._encode_to_be_ticked(
                chunk.misc.get("to_be_ticked"), floor_cy, height_cy
            ),
        )

    def _encode_fluid_ticks(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data,
            self.LiquidTicks,
            self._encode_ticks(chunk.misc.get("fluid_ticks", {})),
        )
        self.set_layer_obj(
            data,
            self.LiquidsToBeTicked,
            self._encode_to_be_ticked(
                chunk.misc.get("liquids_to_be_ticked"), floor_cy, height_cy
            ),
        )

    def _encode_post_processing(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data,
            self.PostProcessing,
            self._encode_to_be_ticked(
                chunk.misc.get("post_processing"), floor_cy, height_cy
            ),
        )

    def _encode_structures(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        self.set_layer_obj(
            data,
            self.Structures,
            chunk.misc.get(
                "structures",
                CompoundTag(
                    {
                        "References": CompoundTag(),
                        "Starts": CompoundTag(),
                    }
                ),
            ),
        )


export = Anvil1444Interface



from typing import Tuple, Sequence
import numpy


def brute_sort_objects(data) -> Tuple[numpy.ndarray, numpy.ndarray]:
    indexes = {}
    unique = []
    inverse = []
    index = 0
    for d in data:
        if d not in indexes:
            indexes[d] = index
            index += 1
            unique.append(d)
        inverse.append(indexes[d])

    unique_ = numpy.empty(len(unique), dtype=object)
    for index, obj in enumerate(unique):
        unique_[index] = obj

    return unique_, numpy.array(inverse)


def brute_sort_objects_no_hash(data) -> Tuple[numpy.ndarray, numpy.ndarray]:
    unique = []
    inverse = numpy.zeros(dtype=numpy.uint32, shape=len(data))
    for i, d in enumerate(data):
        try:
            index = unique.index(d)
        except ValueError:
            index = len(unique)
            unique.append(d)
        inverse[i] = index

    unique_ = numpy.empty(len(unique), dtype=object)
    for index, obj in enumerate(unique):
        unique_[index] = obj

    return unique_, numpy.array(inverse)


def direct_object_array(array: Sequence) -> numpy.ndarray:
    np_array = numpy.empty(len(array), dtype=object)
    for index, obj in array:
        np_array[index] = object
    return np_array



from __future__ import annotations

import sys
import os

from typing import Sequence, Union, TypeVar

_executable_dir = os.path.dirname(sys.executable)
_script_base = os.path.dirname(sys.argv[0])
_package_base = os.path.dirname(os.path.dirname(__file__))
_program_base = os.path.abspath(".")


def _application_directory(
    directory: Union[str, Sequence[str]], src_path: Union[str, Sequence[str]] = None
) -> str:
    """
    Returns a path to a directory that is adjusted depending on whether the program is running in a compiled or as source.
    The necessary directories will be created if they aren't already present.

    If in test mode, then the path won't be created.

    :param directory: The directory or directories to get the full path of and/or create
    :param src_path: Fallback option if the directory location of a compiled build vs. running from source differ
    :return: The full path to the resulting directory
    """
    if isinstance(directory, str):
        directory = (directory,)

    if getattr(sys, "frozen", False):
        path = os.path.join(_executable_dir, *directory)
        if not os.path.exists(path):
            os.makedirs(path)
        return path

    if src_path:
        if isinstance(src_path, str):
            directory = (src_path,)
        else:
            directory = src_path

    if "unittest" in sys.modules:
        return os.path.join(_package_base, *directory)

    else:
        path = os.path.join(_package_base, *directory)
        if not os.path.exists(path):
            os.makedirs(path)
        return path


T = TypeVar("T")


def get_cache_dir(default: T = os.path.join(_program_base, "cache")) -> Union[str, T]:
    return os.environ.get("AMULET_CACHE_DIR", default)



from typing import Optional, Dict
import pkgutil
import importlib
import inspect
import logging

from amulet.api.errors import LoaderNoneMatched
from .base_leveldb_interface import BaseLevelDBInterface

_interfaces: Optional[Dict[int, BaseLevelDBInterface]] = None

log = logging.getLogger(__name__)


def get_interface(interface_version: int):
    global _interfaces
    if _interfaces is None:
        _interfaces = {}

        for _, module_name, _ in pkgutil.walk_packages(__path__, __name__ + "."):
            modu = importlib.import_module(module_name)
            c = getattr(modu, "export", None)
            if inspect.isclass(c) and issubclass(c, BaseLevelDBInterface):
                if c.chunk_version in _interfaces:
                    log.error(
                        f"Multiple Bedrock chunk interfaces with chunk_version={c.chunk_version}"
                    )
                else:
                    _interfaces[c.chunk_version] = c()

    if interface_version not in _interfaces:
        raise LoaderNoneMatched(
            f"Could not find a leveldb interface for chunk version {interface_version}"
        )

    return _interfaces[interface_version]



from typing import overload, Tuple, Union, Optional, Generator
import numpy
import math

from .data_types import (
    SliceSlicesType,
    UnpackedSlicesType,
    DtypeType,
    Integer,
    IntegerType,
)
from .base_partial_3d_array import BasePartial3DArray
from .unbounded_partial_3d_array import UnboundedPartial3DArray
from .util import to_slice, sanitise_slice, unpack_slice, stack_sanitised_slices


class BoundedPartial3DArray(BasePartial3DArray):
    """
    This class should behave the same as a numpy array in all three axis
    but the data internally is stored in sections to minimise memory usage.
    The array has a fixed size in all three axis much like a numpy array.
    """

    @classmethod
    def from_partial_array(
        cls,
        parent_array: UnboundedPartial3DArray,
        start: Tuple[int, int, int],
        stop: Tuple[int, int, int],
        step: Tuple[int, int, int],
    ):
        """
        Create a :class:`BoundedPartial3DArray` from an :class:`~amulet.api.partial_3d_array.UnboundedPartial3DArray` and slices.

        :param parent_array: The :class:`~amulet.api.partial_3d_array.UnboundedPartial3DArray` to "extract" the slice from
        :param start: The starting point of the slice.
        :param stop: The end point of the slice.
        :param step: The steps of the slice.
        :return: A new instance of :class:`BoundedPartial3DArray` that behaves like a view into the parent array.
        """
        return cls(
            parent_array.dtype,
            parent_array.default_value,
            parent_array.section_shape,
            start,
            stop,
            step,
            parent_array,
        )

    def __init__(
        self,
        dtype: DtypeType,
        default_value: Union[int, bool],
        section_shape: Tuple[int, int, int],
        start: Tuple[Optional[int], int, Optional[int]],
        stop: Tuple[Optional[int], int, Optional[int]],
        step: Tuple[Optional[int], Optional[int], Optional[int]],
        parent_array: UnboundedPartial3DArray,
    ):
        """
        Construct a :class:`BoundedPartial3DArray`. This should not be used directly. You should instead use the :meth:`~amulet.api.partial_3d_array.UnboundedPartial3DArray.__getitem__` method of :class:`~amulet.api.partial_3d_array.UnboundedPartial3DArray`

        :param dtype: The dtype that all arrays will be stored in.
        :param default_value: The default value that all undefined arrays will be populated with if required.
        :param section_shape: The shape of each section array.
        :param start: The starting point of the slice.
        :param stop: The end point of the slice.
        :param step: The steps of the slice.
        :param parent_array: The original :class:`~amulet.api.partial_3d_array.UnboundedPartial3DArray` that the slice is viewing into.
        """
        assert isinstance(start[1], int) and isinstance(
            stop[1], int
        ), "start[1] and stop[1] must both be ints."
        assert isinstance(parent_array, UnboundedPartial3DArray)
        super().__init__(
            dtype, default_value, section_shape, start, stop, step, parent_array
        )

    def __repr__(self):
        return f"BoundedPartial3DArray(dtype={self.dtype}, shape={self.shape})"

    def __array__(self, dtype=None):
        """
        Get the data contained within as a numpy array.

        >>> numpy.array(partial_array)

        :param dtype: The dtype of the returned numpy array.
        :return: A numpy array of the contained data.
        """
        array = numpy.full(self.shape, self.default_value, dtype or self.dtype)
        for sy, slices, relative_slices in self._iter_slices(
            (
                (self.start_x, self.stop_x, self.step_x),
                (self.start_y, self.stop_y, self.step_y),
                (self.start_z, self.stop_z, self.step_z),
            )
        ):
            if sy in self._sections:
                array[relative_slices] = self._sections[sy][slices]
        return array

    def __eq__(self, value):
        """
        Check equality of this object and another object.

        Behaves mostly the same as a numpy array.

        >>> bounded_partial_array1 == bounded_partial_array2

        :param value: The object to compare to.
        :return:
        """

        def get_array(default: bool):
            return self.from_partial_array(
                UnboundedPartial3DArray(bool, default, self.section_shape, (0, 0)),
                self.start,
                self.stop,
                self.step,
            )

        if (
            isinstance(value, Integer) and numpy.issubdtype(self.dtype, numpy.integer)
        ) or (isinstance(value, bool) and numpy.issubdtype(self.dtype, bool)):
            out = get_array(value == self.default_value)
            for sy, slices, relative_slices in self._iter_slices(self.slices_tuple):
                if sy in self._sections:
                    out[relative_slices] = self._sections[sy][slices] == value
        elif (
            isinstance(value, (numpy.ndarray, BoundedPartial3DArray))
            and (
                numpy.issubdtype(value.dtype, numpy.integer)
                and numpy.issubdtype(self.dtype, numpy.integer)
            )
            or (
                numpy.issubdtype(value.dtype, bool)
                and numpy.issubdtype(self.dtype, bool)
            )
        ):
            out = get_array(False)
            if self.shape != value.shape:
                raise ValueError(
                    f"The shape of the index ({self.shape}) and the shape of the given array ({value.shape}) do not match."
                )
            for sy, slices, relative_slices in self._iter_slices(self.slices_tuple):
                if sy in self._sections:
                    out[relative_slices] = self._sections[sy][slices] == numpy.asarray(
                        value[relative_slices]
                    )
                else:
                    out[relative_slices] = self.default_value == numpy.asarray(
                        value[relative_slices]
                    )
        else:
            raise ValueError(f"Bad value {value}")
        return out

    def _iter_slices(
        self, slices: UnpackedSlicesType
    ) -> Generator[Tuple[int, SliceSlicesType, SliceSlicesType], None, None]:
        """
        split the sanitised slice into section based slices
        :return: Generator of section y, section slice, relative slice
        """
        slice_x = slice(*slices[0])
        slice_z = slice(*slices[2])
        relative_slice_x = slice(None)
        relative_slice_z = slice(None)

        start_y, stop_y, step_y = slices[1]
        sy = None
        section_start_y = None
        section_stop_y = None
        section_start_dy = None
        section_stop_dy = None
        for y in range(start_y, stop_y, step_y):
            sy_, dy_ = self._section_index(y)
            if sy_ != sy:
                # we are in a new section
                if sy is not None:
                    yield sy, (
                        slice_x,
                        slice(section_start_dy, section_stop_dy, step_y),
                        slice_z,
                    ), (
                        relative_slice_x,
                        slice(
                            math.ceil((section_start_y - start_y) / step_y),
                            math.ceil((section_stop_y - start_y) / step_y),
                        ),
                        relative_slice_z,
                    )
                sy = sy_
                section_start_y = y
                section_start_dy = dy_
            section_stop_y = y + int(math.copysign(1, step_y))
            section_stop_dy = dy_ + int(math.copysign(1, step_y))
        if sy is not None:
            yield sy, (
                slice_x,
                slice(section_start_dy, section_stop_dy, step_y),
                slice_z,
            ), (
                relative_slice_x,
                slice(
                    math.ceil((section_start_y - start_y) / step_y),
                    math.ceil((section_stop_y - start_y) / step_y),
                ),
                relative_slice_z,
            )

    def _relative_to_absolute(self, axis: int, relative_index: int) -> int:
        """Convert a relative index to the absolute value in the array."""
        value = relative_index
        start = self.start[axis]
        stop = self.stop[axis]
        step = self.step[axis]
        if value >= 0:
            value = start + value * step
        else:
            stop_max = start + math.ceil((stop - start) / step) * step
            value = stop_max + value * step

        if step > 0:
            if not start <= value < stop:
                raise IndexError(
                    f"index {relative_index} is out of bounds for axis {axis} with size {self.shape[axis]}"
                )
        else:
            if not start >= value > stop:
                raise IndexError(
                    f"index {relative_index} is out of bounds for axis {axis} with size {self.shape[axis]}"
                )
            value -= 1

        return value

    def _stack_slices(
        self, slices: Tuple[slice, slice, slice]
    ) -> Tuple[Tuple[int, int, int], Tuple[int, int, int], Tuple[int, int, int]]:
        return tuple(
            stack_sanitised_slices(
                start, stop, step, *sanitise_slice(*unpack_slice(to_slice(i)), shape)
            )
            for i, start, stop, step, shape in zip(
                slices, self.start, self.stop, self.step, self.shape
            )
        )

    @overload
    def __getitem__(
        self, slices: Tuple[IntegerType, IntegerType, IntegerType]
    ) -> Union[int, bool]:
        ...

    @overload
    def __getitem__(
        self,
        slices: Tuple[
            Union[IntegerType, slice],
            Union[IntegerType, slice],
            Union[IntegerType, slice],
        ],
    ) -> "BoundedPartial3DArray":
        ...

    @overload
    def __getitem__(
        self, slices: Union[numpy.ndarray, "BoundedPartial3DArray"]
    ) -> numpy.ndarray:
        ...

    def __getitem__(self, item):
        """
        Get a value or sub-section of the unbounded array.

        >>> # get the value at a given location
        >>> value = partial_array[3, 4, 5]  # an integer
        >>> # get a cuboid volume in the array
        >>> value = partial_array[2:3, 4:5, 6:7]  # BoundedPartial3DArray
        >>> # slice and int can be mixed
        >>> value = partial_array[2:3, 4, 6:7]  # BoundedPartial3DArray

        :param item: The slices to extract.
        :return: The value or BoundedPartial3DArray viewing into this array.
        """
        if isinstance(item, tuple):
            if len(item) != 3:
                raise KeyError(f"Tuple item must be of length 3, got {len(item)}")
            if all(isinstance(i, Integer) for i in item):
                x, y, z = tuple(
                    self._relative_to_absolute(axis, item[axis]) for axis in range(3)
                )

                sy, dy = self._section_index(y)
                if sy in self:
                    return int(self._sections[sy][(x, dy, z)])
                else:
                    return self.default_value

            elif all(isinstance(i, (int, numpy.integer, slice)) for i in item):
                item: Tuple[
                    Tuple[int, int, int],
                    Tuple[int, int, int],
                    Tuple[int, int, int],
                ] = zip(*self._stack_slices(item))

                return BoundedPartial3DArray.from_partial_array(
                    self._parent_array, *item
                )
            else:
                raise KeyError(f"Unsupported tuple {item} for getitem")

        elif isinstance(item, (numpy.ndarray, BoundedPartial3DArray)):
            if item.dtype == bool:
                if item.shape != self.shape:
                    raise ValueError(
                        f"The shape of the index ({self.shape}) and the shape of the given array ({item.shape}) do not match."
                    )
                out = []
                for slices_x, relative_slices_x in zip(
                    range(self.start_x, self.stop_x, self.step_x), range(0, self.size_x)
                ):
                    for (
                        sy,
                        (_, slices_y, slices_z),
                        (_, relative_slices_y, relative_slices_z),
                    ) in self._iter_slices(self.slices_tuple):
                        if sy in self._sections:
                            out.append(
                                self._sections[sy][slices_x, slices_y, slices_z][
                                    numpy.asarray(
                                        item[
                                            relative_slices_x,
                                            relative_slices_y,
                                            relative_slices_z,
                                        ]
                                    )
                                ]
                            )
                        else:
                            out.append(
                                numpy.full(
                                    numpy.count_nonzero(
                                        numpy.asarray(
                                            item[
                                                relative_slices_x,
                                                relative_slices_y,
                                                relative_slices_z,
                                            ]
                                        )
                                    ),
                                    self.default_value,
                                    self.dtype,
                                )
                            )
                if out:
                    return numpy.concatenate(out)
                else:
                    return numpy.full(0, self.default_value, self.dtype)
            elif numpy.issubdtype(item.dtype, numpy.integer):
                if isinstance(item, BoundedPartial3DArray):
                    raise ValueError(
                        "Index array with a BoundedPartial3DArray is not valid"
                    )
                raise NotImplementedError(
                    "Index arrays are not currently supported"
                )  # TODO
            else:
                raise ValueError(
                    f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
                )
        else:
            raise KeyError(
                f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
            )

    @overload
    def __setitem__(
        self,
        item: Tuple[
            Union[IntegerType, slice],
            Union[IntegerType, slice],
            Union[IntegerType, slice],
        ],
        value: Union[int, bool, numpy.ndarray, "BoundedPartial3DArray"],
    ):
        ...

    @overload
    def __setitem__(
        self,
        item: Union[numpy.ndarray, "BoundedPartial3DArray"],
        value: Union[int, bool, numpy.ndarray],
    ):
        ...

    def __setitem__(self, item, value):
        """
        Set a sub-section of the array.

        >>> # set the value at a given location
        >>> partial_array[3, 4, 5] = 1
        >>> # set a cuboid volume in the array
        >>> partial_array[2:3, 4:5, 6:7] = 1
        >>> # slice and int can be mixed
        >>> partial_array[2:3, 4, 6:7] = 1

        :param slices: The slices or locations that define the volume to set.
        :param value: The value to set at the given location. Can be an integer or array.
        """
        if isinstance(item, tuple):
            if len(item) != 3:
                raise KeyError(f"Tuple item must be of length 3, got {len(item)}")
            elif all(isinstance(i, (int, numpy.integer, slice)) for i in item):
                stacked_slices: Tuple[
                    Tuple[int, int, int],
                    Tuple[int, int, int],
                    Tuple[int, int, int],
                ] = self._stack_slices(item)
                if (
                    isinstance(value, Integer)
                    and numpy.issubdtype(self.dtype, numpy.integer)
                ) or (isinstance(value, bool) and self.dtype == bool):
                    for sy, slices, _ in self._iter_slices(stacked_slices):
                        if sy in self._sections:
                            self._sections[sy][slices] = value
                        elif value != self.default_value:
                            self._parent_array.create_section(sy)
                            self._sections[sy][slices] = value
                elif (
                    isinstance(value, (numpy.ndarray, BoundedPartial3DArray))
                    and (
                        numpy.issubdtype(value.dtype, numpy.integer)
                        and numpy.issubdtype(self.dtype, numpy.integer)
                    )
                    or (
                        numpy.issubdtype(value.dtype, bool)
                        and numpy.issubdtype(self.dtype, bool)
                    )
                ):
                    size_array = self[item]
                    if size_array.shape != value.shape:
                        raise ValueError(
                            f"The shape of the index ({size_array.shape}) and the shape of the given array ({value.shape}) do not match."
                        )
                    for sy, slices, relative_slices in size_array._iter_slices(
                        stacked_slices
                    ):
                        if sy not in self._sections:
                            self._parent_array.create_section(sy)
                        self._sections[sy][slices] = numpy.asarray(
                            value[relative_slices]
                        )
                else:
                    raise ValueError(f"Bad value {value}")

            else:
                raise KeyError(f"Unsupported tuple {item} for getitem")
        elif isinstance(item, (numpy.ndarray, BoundedPartial3DArray)):
            if item.dtype == bool:
                if item.shape != self.shape:
                    raise ValueError(
                        f"The shape of the index ({self.shape}) and the shape of the given array ({item.shape}) do not match."
                    )
                if isinstance(value, (int, numpy.integer, bool)):
                    for sy, slices, relative_slices in self._iter_slices(
                        self.slices_tuple
                    ):
                        bool_array = numpy.asarray(item[relative_slices])
                        if sy in self._sections:
                            self._sections[sy][slices][bool_array] = value
                        elif value != self.default_value and numpy.any(bool_array):
                            self._parent_array.create_section(sy)
                            self._sections[sy][slices][bool_array] = value
                elif isinstance(value, numpy.ndarray):
                    start = 0
                    true_count = numpy.count_nonzero(item)
                    if true_count != value.size:
                        raise ValueError(
                            f"There are more True values ({true_count}) in the item array than there are values in the value array ({value.size})."
                        )

                    for slices_x, relative_slices_x in zip(
                        range(self.start_x, self.stop_x, self.step_x),
                        range(0, self.size_x),
                    ):
                        for (
                            sy,
                            (_, slices_y, slices_z),
                            (_, relative_slices_y, relative_slices_z),
                        ) in self._iter_slices(self.slices_tuple):
                            if sy not in self._sections:
                                self._parent_array.create_section(sy)
                            bool_array = numpy.asarray(
                                item[
                                    relative_slices_x,
                                    relative_slices_y,
                                    relative_slices_z,
                                ]
                            )
                            count: int = numpy.count_nonzero(bool_array)
                            self._sections[sy][slices_x, slices_y, slices_z][
                                bool_array
                            ] = value[start : start + count]
                            start += count
                else:
                    raise ValueError(
                        f"When setting using a bool array the value must be an int, bool or numpy.ndarray. Got {item.__class__.__name__}({item})"
                    )
            elif numpy.issubdtype(item.dtype, numpy.integer):
                if isinstance(item, BoundedPartial3DArray):
                    raise ValueError(
                        "Index array with a BoundedPartial3DArray is not valid"
                    )
                raise NotImplementedError(
                    "Index arrays are not currently supported"
                )  # TODO
            else:
                raise ValueError(
                    f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
                )
        else:
            raise KeyError(
                f"{item.__class__.__name__}({item}) is not a supported input for __getitem__"
            )



from __future__ import annotations
import os

from amulet_nbt import load as load_nbt

from .anvil_world import AnvilFormat
from amulet.utils.format_utils import check_all_exist


class AnvilForgeFormat(AnvilFormat):
    """
    This FormatWrapper class is a modification on the :class:`AnvilFormat` class that separates Forge worlds from vanilla worlds.

    Currently there is no extra logic here but this should extend the :class:`AnvilFormat` class to support Forge worlds.
    """

    @staticmethod
    def is_valid(path: str) -> bool:
        if not check_all_exist(path, "level.dat"):
            return False

        try:
            level_dat_root = load_nbt(os.path.join(path, "level.dat")).compound
        except:
            return False

        return "Data" in level_dat_root and "FML" in level_dat_root

    @property
    def game_version_string(self) -> str:
        try:
            return f'Java Forge {self.root_tag.compound.get_compound("Data").get_compound("Version").get_string("Name").py_str}'
        except Exception:
            return f"Java Forge Unknown Version"


export = AnvilForgeFormat



from __future__ import annotations

from .anvil_1519 import Anvil1519Interface as ParentInterface


class Anvil1901Interface(ParentInterface):
    """
    Block data in a section is optional
    """

    def __init__(self):
        super().__init__()
        self._unregister_post_encoder(self._post_encode_sections)

    @staticmethod
    def minor_is_valid(key: int):
        return 1901 <= key < 1908


export = Anvil1901Interface



from __future__ import annotations
import struct

from .leveldb_24 import (
    LevelDB24Interface as ParentInterface,
)


class LevelDB25Interface(ParentInterface):
    chunk_version = 25

    @staticmethod
    def _chunk_key_to_sub_chunk(cy: int, min_y: int) -> int:
        return cy + min_y

    @staticmethod
    def _get_sub_chunk_storage_byte(cy: int, min_y: int) -> bytes:
        # The chunk db keys all start at 0 regardless of chunk floor position.
        # This is the floor position of when the world was created.
        # If the floor position changes in the future this will break.
        return struct.pack("b", cy - min_y)


export = LevelDB25Interface



from __future__ import annotations

from typing import Generator, Set, Iterable, Dict

import weakref

from amulet.api.player import Player
from amulet.api.history.history_manager import DatabaseHistoryManager
from amulet.api.history.revision_manager import RAMRevisionManager
from amulet.api import level as api_level
from amulet.api.errors import PlayerLoadError, PlayerDoesNotExist


class PlayerManager(DatabaseHistoryManager):
    _temporary_database: Dict[str, Player]
    _history_database: Dict[str, RAMRevisionManager]

    DoesNotExistError = PlayerDoesNotExist
    LoadError = PlayerLoadError

    def __init__(self, level: api_level.BaseLevel):
        """
        Construct a new :class:`PlayerManager` instance

        This should not be used by third party code

        :param level: The world that this player manager is associated with
        """
        super().__init__()
        self._level = weakref.ref(level)

    @property
    def level(self) -> api_level.BaseLevel:
        """The level that this player manager is associated with."""
        return self._level()

    def all_player_ids(self) -> Set[str]:
        """
        Returns a set of all player ids that are present in the level
        """
        return self._all_entries()

    def _raw_all_entries(self) -> Iterable[str]:
        return self.level.level_wrapper.all_player_ids()

    def changed_players(self) -> Generator[str, None, None]:
        """The player objects that have changed since the last save"""
        return self.changed_entries()

    def has_player(self, player_id: str) -> bool:
        """
        Is the given player id present in the level

        :param player_id: The player id to check
        :return: True if the player id is present, False otherwise
        """
        return self._has_entry(player_id)

    def _raw_has_entry(self, key: str) -> bool:
        return self.level.level_wrapper.has_player(key)

    def __contains__(self, item):
        """
        Is the given player id present in the level

        >>> '<uuid>' in level.players

        :param item: The player id to check
        :return: True if the player id is present, False otherwise
        """
        return self.has_player(item)

    def put_player(self, player: Player):
        """
        Add the given player to the player manager.

        :param player: The :class:`Player` object to add to the chunk manager. It will be added with the key designated by :attr:`Player.uuid`
        """
        self._put_entry(player.player_id, player)

    def get_player(self, player_id: str) -> Player:
        """
        Gets the :class:`Player` object that belongs to the specified player id

        If no parameter is supplied, the data of the local player will be returned

        :param player_id: The desired player id
        :return: A Player instance
        """
        return self._get_entry(player_id)

    def _raw_get_entry(self, key: str) -> Player:
        return self.level.level_wrapper.load_player(key)

    def delete_player(self, player_id: str):
        """
        Deletes a player from the player manager

        :param player_id: The desired player id
        """
        self._delete_entry(player_id)



from __future__ import annotations

from typing import TYPE_CHECKING

from amulet.api.selection import SelectionGroup
from amulet.api.block import Block
from amulet.api.data_types import Dimension, OperationReturnType

if TYPE_CHECKING:
    from amulet.api.level import BaseLevel


def fill(
    world: "BaseLevel",
    dimension: Dimension,
    target_box: SelectionGroup,
    fill_block: Block,
) -> OperationReturnType:
    if not isinstance(fill_block, Block):
        raise Exception("Fill operation was not given a Block object")
    internal_id = world.block_palette.get_add_block(fill_block)

    iter_count = len(list(world.get_coord_box(dimension, target_box, True)))
    count = 0

    for chunk, slices, _ in world.get_chunk_slice_box(dimension, target_box, True):
        chunk.blocks[slices] = internal_id
        chunk.changed = True
        count += 1
        yield count / iter_count



from __future__ import annotations

from amulet.api.wrapper import EntityIDType, EntityCoordType
from .base_leveldb_interface import (
    BaseLevelDBInterface,
)


class LevelDB0Interface(BaseLevelDBInterface):

    chunk_version = 0

    def __init__(self):
        super().__init__()

        self._set_feature("finalised_state", "int0-2")

        self._set_feature("data_2d", "height512|biome256")

        self._set_feature("block_entities", "31list")
        self._set_feature("block_entity_format", EntityIDType.str_id)
        self._set_feature("block_entity_coord_format", EntityCoordType.xyz_int)

        self._set_feature("entities", "32list")
        self._set_feature("entity_format", EntityIDType.int_id)
        self._set_feature("entity_coord_format", EntityCoordType.Pos_list_float)

        self._set_feature("terrain", "30array")


export = LevelDB0Interface



from typing import Generator


def generator_unpacker(gen: Generator):
    """
    Unpack a generator and return the value returned by the generator.

    :param gen: The generator to unpack.
    :return: The value that was returned by the generator.
    """
    try:
        while True:
            next(gen)
    except StopIteration as e:
        return e.value



from __future__ import annotations

from typing import TYPE_CHECKING
from amulet_nbt import StringTag

from amulet.api.chunk import Chunk
from amulet.api.wrapper import Translator
from amulet.api.block import Block
from amulet.api.registry import BlockManager
from amulet.api.data_types import (
    VersionIdentifierType,
    AnyNDArray,
    BlockNDArray,
)

if TYPE_CHECKING:
    from PyMCTranslate import Version, TranslationManager

water = Block.from_string_blockstate("minecraft:water[level=0]")


class JavaBlockstateTranslator(Translator):
    def _translator_key(self, version_number: int) -> VersionIdentifierType:
        return "java", version_number

    @staticmethod
    def is_valid(key):
        return key[0] == "java" and 1444 <= key[1] < 2836

    @staticmethod
    def _unpack_blocks(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
        block_palette: AnyNDArray,
    ):
        """
        Unpack the version-specific block_palette into the stringified version where needed.

        :return: The block_palette converted to block objects.
        """
        version = translation_manager.get_version(*version_identifier)
        for index, block in enumerate(block_palette):
            block: Block
            if version.block.is_waterloggable(block.namespaced_name):
                properties = block.properties
                if "waterlogged" in properties:
                    waterlogged = properties["waterlogged"]
                    del properties["waterlogged"]
                    block = Block(
                        namespace=block.namespace,
                        base_name=block.base_name,
                        properties=properties,
                    )
                else:
                    waterlogged = StringTag("false")

                if waterlogged == StringTag("true"):
                    block_palette[index] = block + water
                else:
                    block_palette[index] = block
            elif version.block.is_waterloggable(block.namespaced_name, True):
                block_palette[index] = block + water

        chunk._block_palette = BlockManager(block_palette)

    def _pack_block_palette(
        self, version: "Version", palette: BlockNDArray
    ) -> AnyNDArray:
        """
        Translate the list of block objects into a version-specific block_palette.
        :return: The block_palette converted into version-specific blocks (ie id, data tuples for 1.12)
        """
        for index, block in enumerate(palette):
            block: Block
            if version.block.is_waterloggable(block.namespaced_name):
                properties = block.properties
                extra_blocks = block.extra_blocks
                if (
                    extra_blocks
                    and extra_blocks[0].namespaced_name == water.namespaced_name
                ):
                    properties["waterlogged"] = StringTag("true")
                else:
                    properties["waterlogged"] = StringTag("false")
                palette[index] = Block(
                    namespace=block.namespace,
                    base_name=block.base_name,
                    properties=properties,
                )
        return palette


export = JavaBlockstateTranslator



from __future__ import annotations

import numpy
from amulet_nbt import IntArrayTag, CompoundTag
from amulet.api.chunk import Chunk
from .base_anvil_interface import ChunkPathType, ChunkDataType
from .anvil_1466 import Anvil1466Interface as ParentInterface


class Anvil1467Interface(ParentInterface):
    """
    Biomes now stored in an int array
    """

    Biomes: ChunkPathType = (
        "region",
        [("Level", CompoundTag), ("Biomes", IntArrayTag)],
        None,
    )

    @staticmethod
    def minor_is_valid(key: int):
        return 1467 <= key < 1484

    def _encode_biomes(
        self, chunk: Chunk, data: ChunkDataType, floor_cy: int, height_cy: int
    ):
        if chunk.status.value > -0.7:
            chunk.biomes.convert_to_2d()
            self.set_layer_obj(
                data,
                self.Biomes,
                IntArrayTag(chunk.biomes.astype(dtype=numpy.uint32).ravel()),
            )


export = Anvil1467Interface



import numpy
from typing import Iterable, Optional, Union, Dict
from copy import deepcopy

from amulet.api.partial_3d_array import UnboundedPartial3DArray


class Blocks(UnboundedPartial3DArray):
    def __init__(
        self,
        input_array: Optional[Union[Dict[int, numpy.ndarray], "Blocks"]] = None,
    ):
        if input_array is None:
            input_array = {}
        if isinstance(input_array, Blocks):
            input_array: dict = deepcopy(input_array._sections)
        if not isinstance(input_array, dict):
            raise Exception(f"Input array must be Blocks or dict, got {input_array}")
        super().__init__(numpy.uint32, 0, (16, 16, 16), (0, 16), sections=input_array)

    @property
    def sub_chunks(self) -> Iterable[int]:
        """An iterable of the sub-chunk indexes that exist"""
        return self.sections

    def has_sub_chunk(self, cy: int) -> bool:
        """Check if the array for a given sub-chunk exists.
        :param cy: The section y index
        :return: True if the array exists, False otherwise
        """
        return self.has_section(cy)

    def get_sub_chunk(self, cy: int) -> numpy.ndarray:
        """Get the section ndarray for a given section index.
        :param cy: The section y index
        :return: Numpy array for this section
        """
        return self.get_section(cy)

    def add_sub_chunk(self, cy: int, sub_chunk: numpy.ndarray):
        """Add a sub-chunk. Overwrite if already exists
        :param cy: The section y index
        :param sub_chunk: The Numpy array to add at this location
        :return:
        """
        self.add_section(cy, sub_chunk)



from collections import UserDict
from typing import Iterable, KeysView, ValuesView, ItemsView
import numpy

from amulet.api.data_types import BlockCoordinates
from amulet.api.block_entity import BlockEntity


class BlockEntityDict(UserDict):
    """
    A custom implementation of the dictionary class.

    It can only store :class:`BlockEntity` instances under the absolute coordinate of the block entity ``Tuple[int, int, int]``
    """

    InputType = Iterable[BlockEntity]

    def __init__(self, block_entities: InputType = ()):
        super().__init__()
        for block_entity in block_entities:
            self._assert_val(block_entity)
            self.data[block_entity.location] = block_entity

    def _assert_key(self, key):
        assert self._check_key(
            key
        ), f"Key must be in the format Tuple[int, int, int]. Got: {key}"

    @staticmethod
    def _check_key(key) -> bool:
        return (
            isinstance(key, tuple)
            and len(key) == 3
            and all(isinstance(a, (int, numpy.integer)) for a in key)
        )

    def _assert_val(self, value):
        assert self._check_val(value), f"Val must be a BlockEntity. Got: {value}"

    @staticmethod
    def _check_val(value) -> bool:
        return isinstance(value, BlockEntity)

    def __repr__(self) -> str:
        super_repr = (
            "".join(f"\n\t{key}:{val}" for key, val in self.data.items()) + "\n"
        )
        return f"BlockEntityDict({super_repr})"

    def clear(self) -> None:
        """Remove all block entities from the chunk."""
        self.data.clear()

    def keys(self) -> KeysView[BlockCoordinates]:
        """The location of every block entity in the chunk. Absolute coordinates."""
        return self.data.keys()

    def __iter__(self) -> Iterable[BlockEntity]:
        """
        An iterable of all the :class:`BlockEntity` objects.
        """
        yield from self.data.values()

    def values(self) -> ValuesView[BlockEntity]:
        """
        An iterable of all the :class:`BlockEntity` objects.
        """
        return self.data.values()

    def items(self) -> ItemsView[BlockCoordinates, BlockEntity]:
        """
        An iterable of all the locations and :class:`BlockEntity` objects.
        """
        return self.data.items()

    def copy(self) -> "BlockEntityDict":
        """
        Create a shallow copy of the block entity container.
        """
        return BlockEntityDict(self.values())

    def insert(self, block_entity: BlockEntity):
        """
        Insert the given :class:`BlockEntity` into the chunk at the location :attr:`BlockEntity.location`

        If a block entity already exists at this location it will be overwritten.

        :param block_entity: The block entity to add to the chunk.
        """
        self._assert_val(block_entity)
        self.data[block_entity.location] = block_entity

    def pop(self, coordinate: BlockCoordinates) -> BlockEntity:
        """
        Remove and return the :class:`BlockEntity` at ``coordinate``.

        :param coordinate: The coordinate to remove the block entity from.
        :return: The block entity at the specified coordinate.
        :raises:
            KeyError if there is no BlockEntity at the given coordinate.
        """
        self._assert_key(coordinate)
        if coordinate in self.data:
            return self.data.pop(coordinate)
        raise KeyError

    def __delitem__(self, coordinate: BlockCoordinates):
        """
        Remove the :class:`BlockEntity` at ``coordinate``.

        :param coordinate: The coordinate to remove the block entity from.
        """
        self._assert_key(coordinate)
        super().__delitem__(coordinate)

    def _check_block_entity(
        self, coordinate: BlockCoordinates, block_entity: BlockEntity
    ) -> BlockEntity:
        self._assert_key(coordinate)
        self._assert_val(block_entity)
        if coordinate != block_entity.location:
            block_entity = block_entity.new_at_location(*coordinate)
        return block_entity

    def __getitem__(self, coordinate: BlockCoordinates) -> BlockEntity:
        """
        Get the :class:`BlockEntity` at ``coordinate``.

        >>> block_entity = chunk.block_entities[(x, y, z)]

        :param coordinate: The coordinate to find the block entity at.
        :return: The block entity at the specified coordinate.
        :raises:
            KeyError if there is no BlockEntity at the given coordinate.
        """
        self._assert_key(coordinate)
        return super().__getitem__(coordinate)

    def __setitem__(self, coordinate: BlockCoordinates, block_entity: BlockEntity):
        """
        Set the :class:`BlockEntity` at ``coordinate``.

        >>> chunk.block_entities[(x, y, z)] = block_entity

        :param coordinate: The coordinate to set the block entity at.
        :param block_entity: The block entity to set at the specified coordinate.
        """
        self.data[coordinate] = self._check_block_entity(coordinate, block_entity)

    def setdefault(
        self, coordinate: BlockCoordinates, block_entity: BlockEntity
    ) -> BlockEntity:
        """
        Set the block entity at the given coordinate if there is not a block entity present.

        :param coordinate: The coordinate to set the block entity at.
        :param block_entity: The block entity to set at the specified coordinate if one is not present.
        """
        block_entity = self._check_block_entity(coordinate, block_entity)
        return self.data.setdefault(coordinate, block_entity)

    def popitem(self):
        raise NotImplementedError

    def update(self, block_entities: InputType):
        """
        Add the sequence of :class:`BlockEntity` instances to the chunk at the location :attr:`BlockEntity.location`

        If multiple block entities have the same coordinate only the last one will remain.

        :param block_entities: A sequence of :class:`BlockEntity` objects to add to the chunk.
        """
        for block_entity in block_entities:
            self._assert_val(block_entity)
            self.data[block_entity.location] = block_entity



from __future__ import annotations

from .anvil_1912 import Anvil1912Interface as ParentInterface


class Anvil1934Interface(ParentInterface):
    """
    Made lighting optional
    """

    def __init__(self):
        super().__init__()
        self._set_feature("light_optional", "true")

    @staticmethod
    def minor_is_valid(key: int):
        return 1934 <= key < 2203


export = Anvil1934Interface



import os
from typing import (
    Optional,
    List,
    Tuple,
    Dict,
    Iterable,
    TYPE_CHECKING,
    BinaryIO,
    Any,
    Union,
)
import numpy
from io import BytesIO
import struct
import copy
import logging

from amulet_nbt import (
    ByteTag,
    IntTag,
    StringTag,
    ListTag,
    CompoundTag,
    ByteArrayTag,
    IntArrayTag,
    NamedTag,
    load as load_nbt,
)


from amulet.api.data_types import (
    AnyNDArray,
    VersionNumberAny,
    VersionNumberTuple,
    Dimension,
    PlatformType,
    ChunkCoordinates,
)
from amulet.api.registry import BlockManager
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ChunkDoesNotExist, ObjectWriteError

from .section import ConstructionSection
from .interface import Construction0Interface, ConstructionInterface
from .util import (
    unpack_palette,
    parse_entities,
    parse_block_entities,
    serialise_entities,
    serialise_block_entities,
    find_fitting_array_type,
    pack_palette,
)

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface

log = logging.getLogger(__name__)

construction_0_interface = Construction0Interface()

INT_STRUCT = struct.Struct(">I")
SECTION_ENTRY_TYPE = numpy.dtype(
    [
        ("sx", "i4"),
        ("sy", "i4"),
        ("sz", "i4"),
        ("shapex", "i1"),
        ("shapey", "i1"),
        ("shapez", "i1"),
        ("position", "i4"),
        ("length", "i4"),
    ]
)

magic_num = b"constrct"
magic_num_len = len(magic_num)

max_format_version = 0
max_section_version = 0


class ConstructionFormatWrapper(StructureFormatWrapper[VersionNumberTuple]):
    """
    This FormatWrapper class exists to interface with the construction format.
    """

    _format_version: int
    _section_version: int
    _chunk_to_section: Dict[Tuple[int, int], List[ConstructionSection]]
    _selection_boxes: List[SelectionBox]
    _chunk_to_box: Dict[Tuple[int, int], List[SelectionBox]]

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`ConstructionFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)

        self._format_version = max_format_version
        self._section_version = max_section_version

        # which sections are in a given chunk
        self._chunk_to_section = {}

        self._selection_boxes = []

        # which selection boxes intersect a given chunk (boxes are clipped to the size of the chunk)
        self._chunk_to_box = {}

        self._shallow_load()

    def _shallow_load(self):
        if os.path.isfile(self.path):
            with open(self.path, "rb") as f:
                magic_num_1 = f.read(magic_num_len)
                if magic_num_1 == magic_num:
                    format_version = struct.unpack(">B", f.read(1))[0]
                    if format_version == 0:
                        f.seek(-magic_num_len, os.SEEK_END)
                        magic_num_2 = f.read(magic_num_len)
                        if magic_num_2 == magic_num:
                            f.seek(-magic_num_len - INT_STRUCT.size, os.SEEK_END)
                            metadata_end = f.tell()
                            metadata_start = INT_STRUCT.unpack(f.read(INT_STRUCT.size))[
                                0
                            ]
                            f.seek(metadata_start)

                            metadata = load_nbt(
                                f.read(metadata_end - metadata_start),
                                compressed=True,
                            ).compound

                            export_version = metadata.get_compound("export_version")

                            self._platform = export_version.get_string("edition").py_str
                            self._version = tuple(
                                map(
                                    lambda v: v.py_int,
                                    export_version.get_list("version"),
                                )
                            )

                            selection_boxes = (
                                metadata.get_int_array("selection_boxes")
                                .np_array.reshape(-1, 6)
                                .tolist()
                            )
                            self._bounds[self.dimensions[0]] = SelectionGroup(
                                [
                                    SelectionBox((minx, miny, minz), (maxx, maxy, maxz))
                                    for minx, miny, minz, maxx, maxy, maxz in selection_boxes
                                ]
                            )

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        format_version=max_format_version,
        section_version=max_section_version,
        **kwargs,
    ):
        if not overwrite and os.path.isfile(self.path):
            raise ObjectWriteError(f"There is already a file at {self.path}")
        self._format_version = format_version
        self._section_version = section_version
        translator_version = self.translation_manager.get_version(
            self.platform, self.version
        )
        self._platform = translator_version.platform
        self._version = translator_version.version_number
        self._chunk_to_section = {}
        self._chunk_to_box = {}
        self._set_selection(bounds)
        self._populate_chunk_to_box()
        self._is_open = True
        self._has_lock = True

    def _populate_chunk_to_box(self):
        for box in self._bounds[self.dimensions[0]].selection_boxes:
            for cx, cz in box.chunk_locations():
                self._chunk_to_box.setdefault((cx, cz), []).append(
                    SelectionBox.create_chunk_box(cx, cz).intersection(box)
                )

    def open_from(self, f: BinaryIO):
        f = BytesIO(f.read())
        magic_num_1 = f.read(magic_num_len)
        assert magic_num_1 == magic_num, f"This file is not a construction file."
        self._format_version = struct.unpack(">B", f.read(1))[0]
        if self._format_version == 0:
            f.seek(-magic_num_len, os.SEEK_END)
            magic_num_2 = f.read(magic_num_len)
            assert (
                magic_num_2 == magic_num
            ), "It looks like this file is corrupt. It probably wasn't saved properly"

            f.seek(-magic_num_len - INT_STRUCT.size, os.SEEK_END)
            metadata_end = f.tell()
            metadata_start = INT_STRUCT.unpack(f.read(INT_STRUCT.size))[0]
            f.seek(metadata_start)

            metadata = load_nbt(
                f.read(metadata_end - metadata_start),
                compressed=True,
            ).compound

            try:
                export_version = metadata.get_compound("export_version")
                self._platform = export_version.get_string("edition").py_str
                self._version = tuple(
                    map(lambda v: v.py_int, export_version.get_list("version"))
                )
            except KeyError as e:
                raise KeyError(f'Missing export version identifying key "{e.args[0]}"')

            self._section_version = metadata.get_byte("section_version").py_int

            palette = unpack_palette(metadata.get_list("block_palette"))

            selection_boxes = (
                metadata.get_int_array("selection_boxes")
                .np_array.reshape(-1, 6)
                .tolist()
            )

            self._bounds[self.dimensions[0]] = SelectionGroup(
                [
                    SelectionBox((minx, miny, minz), (maxx, maxy, maxz))
                    for minx, miny, minz, maxx, maxy, maxz in selection_boxes
                ]
            )

            self._populate_chunk_to_box()

            section_index_table = (
                metadata.get_byte_array("section_index_table")
                .np_array.view(SECTION_ENTRY_TYPE)
                .tolist()
            )

            if self._section_version == 0:
                for (
                    start_x,
                    start_y,
                    start_z,
                    shape_x,
                    shape_y,
                    shape_z,
                    position,
                    length,
                ) in section_index_table:
                    f.seek(position)
                    nbt_obj = load_nbt(f.read(length)).compound
                    blocks_array_type = nbt_obj.get_byte("blocks_array_type").py_int
                    if blocks_array_type == -1:
                        blocks = None
                        block_entities = None
                    else:
                        if blocks_array_type == 7:
                            block_array = nbt_obj.get_byte_array("blocks").np_array
                        elif blocks_array_type == 11:
                            block_array = nbt_obj.get_int_array("blocks").np_array
                        elif blocks_array_type == 12:
                            block_array = nbt_obj.get_long_array("blocks").np_array
                        else:
                            raise TypeError

                        blocks = block_array.reshape((shape_x, shape_y, shape_z))
                        block_entities = parse_block_entities(
                            nbt_obj.get_list("block_entities")
                        )

                    start = numpy.array([start_x, start_y, start_z])
                    chunk_index: numpy.ndarray = start // self.sub_chunk_size
                    shape = numpy.array([shape_x, shape_y, shape_z])
                    if numpy.any(shape <= 0):
                        continue  # skip sections with zero size
                    if numpy.any(
                        start + shape > (chunk_index + 1) * self.sub_chunk_size
                    ):
                        log.error(
                            f"section in construction file did not fit in one sub-chunk. Start: {start}, Shape: {shape}"
                        )
                    cx, cy, cz = chunk_index.tolist()
                    self._chunk_to_section.setdefault((cx, cz), []).append(
                        ConstructionSection(
                            (start_x, start_y, start_z),
                            (shape_x, shape_y, shape_z),
                            blocks,
                            palette,
                            parse_entities(nbt_obj.get_list("entities")),
                            block_entities,
                        )
                    )
            else:
                raise Exception(
                    f"This wrapper does not support any construction section version higher than {max_section_version}"
                )

        else:
            raise Exception(
                f"This wrapper does not support any construction format version higher than {max_format_version}"
            )

    @property
    def multi_selection(self) -> bool:
        return True

    @staticmethod
    def is_valid(path: str) -> bool:
        return os.path.isfile(path) and path.endswith(".construction")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {
            "bedrock": (True, True),
            "java": (True, True),
        }

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".construction",)

    def _get_interface(
        self, raw_chunk_data: Optional[Any] = None
    ) -> Construction0Interface:
        return construction_0_interface

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        palette: BlockManager = BlockManager()
        f.write(magic_num)
        f.write(struct.pack(">B", self._format_version))
        if self._format_version == 0:
            metadata = CompoundTag(
                {
                    "created_with": StringTag("amulet_python_wrapper_v2"),
                    "selection_boxes": IntArrayTag(
                        [
                            c
                            for box in self._bounds[self.dimensions[0]].selection_boxes
                            for c in (*box.min, *box.max)
                        ]
                    ),
                    "section_version": ByteTag(self._section_version),
                    "export_version": CompoundTag(
                        {
                            "edition": StringTag(self._platform),
                            "version": ListTag([IntTag(v) for v in self._version]),
                        }
                    ),
                }
            )
            section_index_table: List[
                Tuple[int, int, int, int, int, int, int, int]
            ] = []
            if self._section_version == 0:
                for section_list in self._chunk_to_section.values():
                    for section in section_list:
                        sx, sy, sz = section.location
                        shapex, shapey, shapez = section.shape
                        blocks = section.blocks
                        entities = section.entities
                        block_entities = section.block_entities
                        section_palette = section.palette
                        position = f.tell()

                        _tag = CompoundTag({"entities": serialise_entities(entities)})

                        if blocks is None:
                            _tag["blocks_array_type"] = ByteTag(-1)
                        else:
                            flattened_array = blocks.ravel()
                            index, flattened_array = numpy.unique(
                                flattened_array, return_inverse=True
                            )
                            section_palette = numpy.array(
                                section_palette, dtype=object
                            )[index]
                            lut = numpy.vectorize(palette.get_add_block)(
                                section_palette
                            )
                            flattened_array = lut[flattened_array]
                            array_type = find_fitting_array_type(flattened_array)
                            _tag["blocks_array_type"] = ByteTag(array_type.tag_id)
                            _tag["blocks"] = array_type(flattened_array)
                            _tag["block_entities"] = serialise_block_entities(
                                block_entities or []
                            )

                        NamedTag(_tag).save_to(f)

                        length = f.tell() - position
                        section_index_table.append(
                            (sx, sy, sz, shapex, shapey, shapez, position, length)
                        )
            else:
                raise Exception(
                    f"This wrapper doesn't support any section version higher than {max_section_version}"
                )
            metadata_start = f.tell()
            metadata["section_index_table"] = ByteArrayTag(
                numpy.array(section_index_table, dtype=SECTION_ENTRY_TYPE).view(
                    numpy.int8
                )
            )
            metadata["block_palette"] = pack_palette(palette)
            metadata.save_to(f)
            f.write(INT_STRUCT.pack(metadata_start))
            f.write(magic_num)
        else:
            raise Exception(
                f"This wrapper doesn't support any construction version higher than {max_format_version}"
            )

    def _close(self):
        """Close the disk database"""
        pass

    def unload(self):
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Iterable[ChunkCoordinates]:
        yield from self._chunk_to_section.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunk_to_section

    def _pack(
        self,
        chunk: "Chunk",
        translator: "Translator",
        chunk_version: VersionNumberAny,
    ) -> Tuple["Chunk", AnyNDArray]:
        return chunk, numpy.array(chunk.block_palette.blocks)

    def _encode(
        self,
        interface: ConstructionInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            self._chunk_to_box.get((chunk.cx, chunk.cz)),
        )

    def _unpack(
        self,
        translator: "Translator",
        game_version: VersionNumberAny,
        chunk: "Chunk",
        chunk_palette: AnyNDArray,
    ) -> "Chunk":
        palette = chunk._block_palette = BlockManager()
        lut = numpy.array([palette.get_add_block(block) for block in chunk_palette])
        if len(palette.blocks) != len(chunk_palette):
            # if a blockstate was defined twice
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, lut[chunk.blocks.get_sub_chunk(cy)])
        return chunk

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunk_to_section:
            del self._chunk_to_section[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        data: List[ConstructionSection],
        dimension: Optional[Dimension] = None,
    ):
        self._chunk_to_section[cx, cz] = copy.deepcopy(data)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> List[ConstructionSection]:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunk_to_section:
            return copy.deepcopy(self._chunk_to_section[(cx, cz)])
        else:
            raise ChunkDoesNotExist



import os
from typing import Optional, Tuple, Iterable, TYPE_CHECKING, BinaryIO, Dict, List, Union
import numpy
import copy

from amulet_nbt import (
    ShortTag,
    IntTag,
    ListTag,
    CompoundTag,
    ByteArrayTag,
    IntArrayTag,
    NamedTag,
    load as load_nbt,
)

from amulet.api.data_types import (
    VersionNumberAny,
    VersionNumberInt,
    ChunkCoordinates,
    AnyNDArray,
    Dimension,
    PlatformType,
)
from amulet.api.wrapper import StructureFormatWrapper
from amulet.api.chunk import Chunk
from amulet.api.selection import SelectionGroup, SelectionBox
from amulet.api.errors import ChunkDoesNotExist, ObjectWriteError, ObjectReadError
from amulet.api.block import Block
from amulet.utils.numpy_helpers import brute_sort_objects_no_hash

from .chunk import SpongeSchemChunk
from .interface import SpongeSchemInterface
from .varint import decode_byte_array, encode_array

if TYPE_CHECKING:
    from amulet.api.wrapper import Translator, Interface


class SpongeSchemReadError(ObjectReadError):
    pass


class SpongeSchemWriteError(ObjectWriteError):
    pass


sponge_schem_interface = SpongeSchemInterface()

max_schem_version = 2


class SpongeSchemFormatWrapper(StructureFormatWrapper[VersionNumberInt]):
    """
    This FormatWrapper class exists to interface with the sponge schematic structure format.
    """

    def __init__(self, path: str):
        """
        Construct a new instance of :class:`SpongeSchemFormatWrapper`.

        This should not be used directly. You should instead use :func:`amulet.load_format`.

        :param path: The file path to the serialised data.
        """
        super().__init__(path)
        self._chunks: Dict[
            ChunkCoordinates,
            SpongeSchemChunk,
        ] = {}
        self._schem_version: int = max_schem_version

    def _create(
        self,
        overwrite: bool,
        bounds: Union[
            SelectionGroup, Dict[Dimension, Optional[SelectionGroup]], None
        ] = None,
        **kwargs,
    ):
        if not overwrite and os.path.isfile(self.path):
            raise SpongeSchemWriteError(f"There is already a file at {self.path}")
        translator_version = self.translation_manager.get_version("java", self._version)
        self._platform = translator_version.platform
        self._version = translator_version.data_version
        self._chunks = {}
        self._set_selection(bounds)
        self._is_open = True
        self._has_lock = True

    def open_from(self, f: BinaryIO):
        sponge_schem = load_nbt(f).compound
        version_tag = sponge_schem.get("Version")
        if not isinstance(version_tag, IntTag):
            raise SpongeSchemReadError("Version key must exist and be an integer.")
        version = version_tag.py_int
        if version == 1:
            raise SpongeSchemReadError(
                "Sponge Schematic Version 1 is not supported currently."
            )
        elif version == 2:
            offset = sponge_schem.get("Offset")
            if isinstance(offset, IntArrayTag) and len(offset) == 3:
                min_point = numpy.array(offset)
            else:
                min_point = numpy.array([0, 0, 0], dtype=numpy.int32)

            size = []
            for key in ("Width", "Height", "Length"):
                val = sponge_schem.get(key)
                if not isinstance(val, ShortTag):
                    raise SpongeSchemReadError(
                        f"Key {key} must exist and be a ShortTag."
                    )
                # convert to an unsigned short
                val = val.py_int
                if val < 0:
                    val += 2**16
                size.append(val)

            max_point = min_point + size
            selection = SelectionBox(min_point, max_point)
            self._bounds[self.dimensions[0]] = SelectionGroup(selection)
            data_version = sponge_schem.get("DataVersion")
            if not isinstance(data_version, IntTag):
                raise SpongeSchemReadError("DataVersion must be a IntTag.")
            translator_version = self.translation_manager.get_version(
                "java", int(data_version)
            )
            self._platform = translator_version.platform
            self._version = translator_version.data_version

            packed_block_data = sponge_schem.get("BlockData")
            if not isinstance(packed_block_data, ByteArrayTag):
                raise SpongeSchemReadError("BlockData must be a ByteArrayTag")

            unpacked_block_data = decode_byte_array(
                numpy.array(packed_block_data, dtype=numpy.uint8)
            )
            if len(unpacked_block_data) != numpy.prod(size):
                raise SpongeSchemReadError(
                    "The data contained in BlockData does not match the size of the schematic."
                )
            dx, dy, dz = selection.shape
            blocks_array: numpy.ndarray = numpy.transpose(
                numpy.array(
                    unpacked_block_data,
                    dtype=numpy.uint32,
                ).reshape((dy, dz, dx)),
                (2, 0, 1),  # YZX => XYZ
            )

            if "Palette" not in sponge_schem:
                raise SpongeSchemReadError(
                    "Amulet is not able to read Sponge Schem files with no block palette."
                )

            palette_data = sponge_schem.get("Palette")
            if not isinstance(palette_data, CompoundTag):
                raise SpongeSchemReadError("Palette must be a CompoundTag.")

            block_palette: Dict[int, Block] = {}
            for blockstate, index_tag in palette_data.items():
                index = index_tag.py_int
                if index in block_palette:
                    raise SpongeSchemReadError(
                        f"Duplicate block index {index} found in the palette."
                    )
                block_palette[index] = Block.from_string_blockstate(blockstate)

            if not numpy.all(numpy.isin(blocks_array, list(block_palette))):
                raise SpongeSchemReadError(
                    "Some values in BlockData were not present in Palette"
                )

            for cx, cz in selection.chunk_locations():
                chunk_box = SelectionBox.create_chunk_box(cx, cz).intersection(
                    selection
                )
                array_slice = chunk_box.create_moved_box(
                    selection.min, subtract=True
                ).slice
                chunk_blocks_: numpy.ndarray = blocks_array[array_slice]
                chunk_palette_indexes, chunk_blocks = numpy.unique(
                    chunk_blocks_,
                    return_inverse=True,
                )
                chunk_blocks = chunk_blocks.reshape(chunk_blocks_.shape)

                chunk_palette = numpy.empty(len(chunk_palette_indexes), dtype=object)
                for palette_index, index in enumerate(chunk_palette_indexes):
                    chunk_palette[palette_index] = block_palette[index]

                self._chunks[(cx, cz)] = SpongeSchemChunk(
                    chunk_box,
                    chunk_blocks,
                    chunk_palette,
                    [],
                    [],
                )

            if "BlockEntities" in sponge_schem:
                block_entities = sponge_schem["BlockEntities"]
                if (
                    not isinstance(block_entities, ListTag)
                    or block_entities.list_data_type != 10  # CompoundTag.tag_id
                ):
                    raise SpongeSchemReadError(
                        "BlockEntities must be a ListTag of compound tags."
                    )

                for block_entity in block_entities:
                    if "Pos" not in block_entity:
                        continue

                    pos_tag = block_entity["Pos"]
                    if not (isinstance(pos_tag, IntArrayTag) and len(pos_tag) == 3):
                        continue

                    pos = pos_tag.np_array + min_point
                    x, y, z = pos
                    block_entity["Pos"] = IntArrayTag(pos)
                    cx, cz = x >> 4, z >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[
                        (cx, cz)
                    ].selection:
                        self._chunks[(cx, cz)].block_entities.append(block_entity)

            if "Entities" in sponge_schem:
                entities = sponge_schem["Entities"]
                if (
                    not isinstance(entities, ListTag)
                    or entities.list_data_type != 10  # CompoundTag.tag_id
                ):
                    raise SpongeSchemReadError(
                        "Entities must be a ListTag of compound tags."
                    )

                for entity in entities:
                    if "Pos" not in entity:
                        continue

                    pos = entity["Pos"]
                    if not (
                        isinstance(pos, ListTag)
                        and len(pos) == 3
                        and pos.list_data_type == 6
                    ):  # DoubleTag.tag_id:
                        continue

                    x, y, z = (
                        pos[0].py_float + offset[0],
                        pos[1].py_float + offset[0],
                        pos[2].py_float + offset[0],
                    )
                    entity["Pos"] = ListTag(
                        [
                            IntTag(x),
                            IntTag(y),
                            IntTag(z),
                        ]
                    )
                    cx, cz = numpy.floor([x, z]).astype(int) >> 4
                    if (cx, cz) in self._chunks and (x, y, z) in self._chunks[
                        (cx, cz)
                    ].selection:
                        self._chunks[(cx, cz)].entities.append(entity)

        else:
            raise SpongeSchemReadError(
                f"Sponge Schematic Version {version} is not supported currently."
            )

    @staticmethod
    def is_valid(path: str) -> bool:
        return os.path.isfile(path) and path.endswith(".schem")

    @property
    def valid_formats(self) -> Dict[PlatformType, Tuple[bool, bool]]:
        return {"java": (False, True)}

    @property
    def extensions(self) -> Tuple[str, ...]:
        return (".schem",)

    def _get_interface(self, raw_chunk_data=None) -> "SpongeSchemInterface":
        return sponge_schem_interface

    def _get_interface_and_translator(
        self, raw_chunk_data=None
    ) -> Tuple["Interface", "Translator", VersionNumberAny]:
        interface = self._get_interface(raw_chunk_data)
        translator, version_identifier = interface.get_translator(
            self.max_world_version, raw_chunk_data, self.translation_manager
        )
        return interface, translator, version_identifier

    def save_to(self, f: BinaryIO):
        if self._schem_version == 1:
            raise SpongeSchemReadError(
                "Sponge Schematic Version 1 is not supported currently."
            )
        elif self._schem_version == 2:
            selection = self._bounds[self.dimensions[0]].selection_boxes[0]
            if any(s > 2**16 - 1 for s in selection.shape):
                raise SpongeSchemWriteError(
                    "The structure is too large to be exported to a Sponge Schematic file. It must be 2^16 - 1 at most in each dimension."
                )
            overflowed_shape = [
                s if s < 2**15 else s - 2**16 for s in selection.shape
            ]
            tag = CompoundTag(
                {
                    "Version": IntTag(2),
                    "DataVersion": IntTag(self._version),
                    "Width": ShortTag(overflowed_shape[0]),
                    "Height": ShortTag(overflowed_shape[1]),
                    "Length": ShortTag(overflowed_shape[2]),
                    "Offset": IntArrayTag(selection.min),
                }
            )

            entities = []
            block_entities = []
            blocks = numpy.zeros(selection.shape, dtype=numpy.uint32)
            palette: List[AnyNDArray] = []
            if self._version < 1500:
                raise Exception(
                    "Writing to Sponge Schematic files in pre-1.13 format is not currently supported."
                )
            else:
                arr = numpy.empty(1, dtype=object)
                arr[0] = Block("minecraft", "air")
                palette.append(arr)

            palette_len = 1

            for chunk in self._chunks.values():
                if chunk.selection.intersects(selection):
                    box = chunk.selection.create_moved_box(selection.min, subtract=True)
                    blocks[box.slice] = chunk.blocks + palette_len
                    palette.append(chunk.palette)
                    palette_len += len(chunk.palette)
                    for be in chunk.block_entities:
                        be = copy.deepcopy(be)
                        be["Pos"] = IntArrayTag(be["Pos"].np_array - selection.min)
                        block_entities.append(be)

                    for e in chunk.entities:
                        e = copy.deepcopy(e)
                        x, y, z = e["Pos"]
                        e["Pos"] = ListTag(
                            [
                                IntTag(x - selection.min_x),
                                IntTag(y - selection.min_y),
                                IntTag(z - selection.min_z),
                            ]
                        )
                        entities.append(e)

            compact_palette, lut = brute_sort_objects_no_hash(
                numpy.concatenate(palette)
            )
            blocks = numpy.transpose(lut[blocks], (1, 2, 0)).ravel()  # XYZ => YZX
            block_palette = []
            for index, block in enumerate(compact_palette):
                block: Block
                block_palette.append(block.blockstate)

            tag["PaletteMax"] = IntTag(len(compact_palette))
            tag["Palette"] = CompoundTag(
                {
                    blockstate: IntTag(index)
                    for index, blockstate in enumerate(block_palette)
                }
            )
            tag["BlockData"] = ByteArrayTag(list(encode_array(blocks)))
            if block_entities:
                tag["BlockEntities"] = ListTag(block_entities)
            if entities:
                tag["Entities"] = ListTag(entities)

            NamedTag(tag, "Schematic").save_to(f)
        else:
            raise SpongeSchemReadError(
                f"Sponge Schematic Version {self._schem_version} is not supported currently."
            )

    def _close(self):
        """Close the disk database"""
        self._chunks.clear()

    def unload(self):
        pass

    def all_chunk_coords(
        self, dimension: Optional[Dimension] = None
    ) -> Iterable[ChunkCoordinates]:
        yield from self._chunks.keys()

    def has_chunk(self, cx: int, cz: int, dimension: Dimension) -> bool:
        return (cx, cz) in self._chunks

    def _encode(
        self,
        interface: SpongeSchemInterface,
        chunk: Chunk,
        dimension: Dimension,
        chunk_palette: AnyNDArray,
    ):
        return interface.encode(
            chunk,
            chunk_palette,
            self.max_world_version,
            SelectionBox.create_chunk_box(chunk.cx, chunk.cz).intersection(
                self._bounds[dimension].to_box()
            ),
        )

    def _delete_chunk(self, cx: int, cz: int, dimension: Optional[Dimension] = None):
        if (cx, cz) in self._chunks:
            del self._chunks[(cx, cz)]

    def _put_raw_chunk_data(
        self,
        cx: int,
        cz: int,
        section: SpongeSchemChunk,
        dimension: Optional[Dimension] = None,
    ):
        self._chunks[(cx, cz)] = copy.deepcopy(section)

    def _get_raw_chunk_data(
        self, cx: int, cz: int, dimension: Optional[Dimension] = None
    ) -> SpongeSchemChunk:
        """
        Return the raw data as loaded from disk.

        :param cx: The x coordinate of the chunk.
        :param cz: The z coordinate of the chunk.
        :param dimension: The dimension to load the data from.
        :return: The raw chunk data.
        """
        if (cx, cz) in self._chunks:
            return copy.deepcopy(self._chunks[(cx, cz)])
        else:
            raise ChunkDoesNotExist



from abc import ABC, abstractmethod
from typing import Iterable, BinaryIO
import os
import json
from zipfile import ZipFile
import re


class BaseWrapper(ABC):
    def __init__(self, path: str):
        self._path = path

    @staticmethod
    @abstractmethod
    def is_valid(path: str) -> bool:
        """
        Is the given path valid for this class.

        :param path: The path to test.
        """

    @property
    def path(self) -> str:
        """The path to the data."""
        return self._path

    @property
    @abstractmethod
    def all_files(self) -> Iterable[str]:
        """
        The relative paths of all files contained within.

        :return: An iterable of paths.
        """
        raise NotImplementedError

    def all_files_match(self, match: str) -> Iterable[str]:
        """
        The relative paths of all files contained within that match the given regex.

        :param match: The regex string to match again.
        :return: An iterable of file paths that match the given regex.
        """
        re_match = re.compile(match)
        return tuple(path for path in self.all_files if re_match.fullmatch(path))

    @abstractmethod
    def has_file(self, relative_path: str) -> bool:
        """
        Does the requested file exist.

        :param relative_path:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def open(self, relative_path: str, **kwargs) -> BinaryIO:
        """
        Get the contents of the file.

        :param relative_path:
        :return:
        """
        raise NotImplementedError

    @abstractmethod
    def close(self):
        """Close the contents."""
        raise NotImplementedError


class ZipWrapper(BaseWrapper, ZipFile):
    def __init__(self, path: str):
        BaseWrapper.__init__(self, path)
        ZipFile.__init__(self, path)

    @staticmethod
    def is_valid(path: str) -> bool:
        return os.path.isfile(path) and path.endswith(".zip")

    @property
    def all_files(self) -> Iterable[str]:
        for path in self.NameToInfo:
            if not path.endswith("/"):
                yield path

    def has_file(self, relative_path: str) -> bool:
        if relative_path.endswith("/"):
            # is a directory
            return False
        else:
            return relative_path in self.NameToInfo

    open = ZipFile.open
    close = ZipFile.close


class DirWrapper(BaseWrapper):
    def __init__(self, path: str):
        BaseWrapper.__init__(self, path)

    @staticmethod
    def is_valid(path: str) -> bool:
        return os.path.isdir(path)

    @property
    def all_files(self) -> Iterable[str]:
        for abs_path, _, files in os.walk(self.path):
            rel_path = os.path.normpath(os.path.relpath(abs_path, self.path)).replace(
                os.sep, "/"
            )
            if rel_path == ".":
                yield from files
            else:
                for f in files:
                    yield f"{rel_path}/{f}"

    def has_file(self, relative_path: str) -> bool:
        return os.path.isfile(os.path.join(self.path, relative_path))

    def open(self, relative_path: str, **kwargs) -> BinaryIO:
        return open(os.path.join(self.path, relative_path), "rb", **kwargs)

    def close(self):
        pass


def _open_wrapper(path: str) -> BaseWrapper:
    for cls in (ZipWrapper, DirWrapper):
        if cls.is_valid(path):
            return cls(path)
    raise FileNotFoundError(f"The given path {path} is not valid.")


class DataPack:
    """The DataPack class wraps a single data pack."""

    def __init__(self, path: str):
        self._path = path
        self._wrapper = _open_wrapper(path)
        self._is_valid = self.is_wrapper_valid(self._wrapper)

    @property
    def path(self) -> str:
        """The path to the data."""
        return self._path

    @staticmethod
    def is_path_valid(path: str) -> bool:
        """
        Check if the given path is a valid data pack.

        :param path: The path to the data pack. Can be a zip file or directory.
        :return: True if the path is a valid data pack, False otherwise.
        """
        try:
            wrapper = _open_wrapper(path)
        except FileNotFoundError:
            return False
        is_valid = DataPack.is_wrapper_valid(wrapper)
        wrapper.close()
        return is_valid

    @property
    def is_valid(self) -> bool:
        return self._is_valid

    @staticmethod
    def is_wrapper_valid(wrapper: BaseWrapper):
        if wrapper.has_file("pack.mcmeta"):
            with wrapper.open("pack.mcmeta") as m:
                try:
                    meta_file = json.load(m)
                except json.JSONDecodeError:
                    pass
                else:
                    if isinstance(meta_file, dict) and isinstance(
                        meta_file.get("pack", {}).get("pack_format", None), int
                    ):
                        # TODO: check the actual value
                        return True
        return False

    @property
    def all_files(self) -> Iterable[str]:
        """
        The relative paths of all files contained within.

        :return: An iterable of paths.
        """
        return self._wrapper.all_files

    def all_files_match(self, match: str) -> Iterable[str]:
        """
        The relative paths of all files contained within that match the given regex.

        :param match: The regex string to match again.
        :return: An iterable of file paths that match the given regex.
        """
        return self._wrapper.all_files_match(match)

    def has_file(self, relative_path: str) -> bool:
        """
        Does the requested file exist.

        :param relative_path:
        :return:
        """
        return self._wrapper.has_file(relative_path)

    def open(self, relative_path: str) -> BinaryIO:
        """
        Get the contents of the file.

        :param relative_path:
        :return:
        """
        return self._wrapper.open(relative_path)

    @abstractmethod
    def close(self):
        """Close the contents."""
        self._wrapper.close()



from __future__ import annotations

from typing import Optional, TYPE_CHECKING
import logging
import numpy

from amulet_nbt import IntTag


from amulet.api.block import Block
from amulet.api.registry import BlockManager
from amulet.api.entity import Entity
from amulet.api.wrapper.chunk.translator import Translator
from amulet.api.data_types import (
    GetBlockCallback,
    TranslateBlockCallbackReturn,
    TranslateEntityCallbackReturn,
    VersionNumberTuple,
    GetChunkCallback,
    BedrockInterfaceBlockType,
    VersionIdentifierType,
    VersionNumberAny,
    BlockCoordinates,
    AnyNDArray,
)

if TYPE_CHECKING:
    from PyMCTranslate import TranslationManager
    from amulet.api.chunk import Chunk

log = logging.getLogger(__name__)


class BaseBedrockTranslator(Translator):
    def _translator_key(
        self, version_number: VersionNumberAny
    ) -> VersionIdentifierType:
        return "bedrock", version_number

    @staticmethod
    def _unpack_blocks(
        translation_manager: "TranslationManager",
        version_identifier: VersionIdentifierType,
        chunk: Chunk,
        block_palette: AnyNDArray,
    ):
        """
        Unpacks an object array of block data into a numpy object array containing Block objects.
        :param translation_manager:
        :param version_identifier:
        :param chunk:
        :param block_palette:
        :type block_palette: numpy.ndarray[
            Tuple[
                Union[
                    Tuple[None, Tuple[int, int]],
                    Tuple[None, Block],
                    Tuple[int, Block]
                ], ...
            ]
        ]
        :return:
        """
        palette_ = BlockManager()
        lut = []
        for palette_index, entry in enumerate(block_palette):
            entry: BedrockInterfaceBlockType
            block = None
            for version_number, b in entry:
                version_number: Optional[int]
                if isinstance(b, tuple):
                    version = translation_manager.get_version(
                        version_identifier[0], version_number or 17563649
                    )
                    b = version.block.ints_to_block(*b)
                elif isinstance(b, Block):
                    if version_number is not None:
                        properties = b.properties
                        properties["__version__"] = IntTag(version_number)
                        b = Block(b.namespace, b.base_name, properties, b.extra_blocks)
                else:
                    raise Exception(f"Unsupported type {b}")
                if block is None:
                    block = b
                else:
                    block += b
            if block is None:
                raise Exception(f"Empty tuple")

            lut.append(palette_.get_add_block(block))
        chunk._block_palette = palette_

        if len(palette_) != len(lut):
            # sometimes a block can be stored in different formats but unpack to the same block
            # this means that the final palette is smaller than the original so the array needs remapping
            np_lut = numpy.array(lut)
            for cy in chunk.blocks.sub_chunks:
                chunk.blocks.add_sub_chunk(cy, np_lut[chunk.blocks.get_sub_chunk(cy)])

    def _blocks_entities_to_universal(
        self,
        game_version: VersionNumberTuple,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ):
        # Bedrock does versioning by block rather than by chunk.
        # As such we can't just pass in a single translator.
        # It needs to be done dynamically.
        versions = {}

        def translate_block(
            input_object: Block,
            get_block_callback: Optional[GetBlockCallback],
            block_location: BlockCoordinates,
        ) -> TranslateBlockCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            final_extra = False

            for depth, block in enumerate(input_object.block_tuple):
                if "__version__" in block.properties:
                    game_version_: int = int(block.properties.get("__version__"))
                else:
                    if "block_data" in block.properties:
                        # if block_data is in properties cap out at 1.12.x
                        game_version_: VersionNumberTuple = min(
                            game_version, (1, 12, 999)
                        )
                    else:
                        game_version_: VersionNumberTuple = game_version
                version_key = self._translator_key(game_version_)
                if version_key not in versions:
                    versions[version_key] = translation_manager.get_version(
                        *version_key
                    ).block.to_universal
                output_object, output_block_entity, extra = versions[version_key](
                    block,
                    get_block_callback=get_block_callback,
                    block_location=block_location,
                )

                if isinstance(output_object, Block):
                    if not output_object.namespace.startswith("universal"):
                        log.debug(
                            f"Error translating {block.full_blockstate} to universal. Got {output_object.full_blockstate}"
                        )
                    if final_block is None:
                        final_block = output_object
                    else:
                        final_block += output_object
                    if depth == 0:
                        final_block_entity = output_block_entity

                elif isinstance(output_object, Entity):
                    final_entities.append(output_object)
                    # TODO: offset entity coords

                final_extra |= extra

            return final_block, final_block_entity, final_entities, final_extra

        def translate_entity(input_object: Entity) -> TranslateEntityCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            # TODO
            return final_block, final_block_entity, final_entities

        version = translation_manager.get_version(*self._translator_key(game_version))

        self._translate(
            chunk,
            get_chunk_callback,
            translate_block,
            translate_entity,
            full_translate,
        )

    def _blocks_entities_from_universal(
        self,
        max_world_version_number: VersionNumberAny,
        translation_manager: "TranslationManager",
        chunk: Chunk,
        get_chunk_callback: Optional[GetChunkCallback],
        full_translate: bool,
    ):
        """
        Translate a universal chunk into the interface-specific format.

        :param max_world_version_number: The version number (int or tuple) of the max world version
        :param translation_manager: TranslationManager used for the translation
        :param chunk: The chunk to translate.
        :param get_chunk_callback: function callback to get a chunk's data
        :param full_translate: if true do a full translate. If false just pack the block_palette (used in callback)
        :return: Chunk object in the interface-specific format and block_palette.
        """
        version = translation_manager.get_version(
            *self._translator_key(max_world_version_number)
        )

        # TODO: perhaps find a way so this code isn't duplicated in three places
        def translate_block(
            input_object: Block,
            get_block_callback: Optional[GetBlockCallback],
            block_location: BlockCoordinates,
        ) -> TranslateBlockCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            final_extra = False

            for depth, block in enumerate(input_object.block_tuple):
                (
                    output_object,
                    output_block_entity,
                    extra,
                ) = version.block.from_universal(
                    block,
                    get_block_callback=get_block_callback,
                    block_location=block_location,
                )

                if isinstance(output_object, Block):
                    if __debug__ and output_object.namespace.startswith("universal"):
                        log.debug(
                            f"Error translating {input_object.full_blockstate} from universal. Got {output_object.full_blockstate}"
                        )
                    if version.data_version > 0:
                        properties = output_object.properties
                        properties["__version__"] = IntTag(version.data_version)
                        output_object = Block(
                            output_object.namespace,
                            output_object.base_name,
                            properties,
                            output_object.extra_blocks,
                        )
                    if final_block is None:
                        final_block = output_object
                    else:
                        final_block += output_object
                    if depth == 0:
                        final_block_entity = output_block_entity

                elif isinstance(output_object, Entity):
                    final_entities.append(output_object)
                    # TODO: offset entity coords

                final_extra |= extra

            return final_block, final_block_entity, final_entities, final_extra

        def translate_entity(input_object: Entity) -> TranslateEntityCallbackReturn:
            final_block = None
            final_block_entity = None
            final_entities = []
            # TODO
            return final_block, final_block_entity, final_entities

        self._translate(
            chunk,
            get_chunk_callback,
            translate_block,
            translate_entity,
            full_translate,
        )



from setuptools import setup, find_packages
from Cython.Build import cythonize
import glob
import versioneer
import numpy

ext = []
pyx_path = f"amulet/**/*.pyx"
if next(glob.iglob(pyx_path, recursive=True), None):
    # This throws an error if it does not match any files
    ext += cythonize(
        pyx_path,
        language_level=3,
        annotate=True,
    )

setup(
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    include_dirs=[numpy.get_include()],
    packages=find_packages(),
    ext_modules=ext,
)



from __future__ import annotations

from typing import List, Optional

from amulet.api.data_types import BlockCoordinates
from amulet import Block
from amulet.api.entity import Entity
from amulet.api.block_entity import BlockEntity

import numpy


class ConstructionSection:
    __slots__ = (
        "sx",
        "sy",
        "sz",
        "blocks",
        "palette",
        "shape",
        "entities",
        "block_entities",
    )

    def __init__(
        self,
        min_position: BlockCoordinates,
        shape: BlockCoordinates,
        blocks: Optional[numpy.ndarray],
        palette: List[Block],
        entities: List[Entity],
        block_entities: List[BlockEntity],
    ):
        self.sx, self.sy, self.sz = min_position
        self.shape = shape
        self.blocks = blocks
        if blocks is not None:
            assert isinstance(self.blocks, numpy.ndarray)
            assert (
                self.shape == self.blocks.shape
            ), "blocks shape does not match the specified section shape"
        self.palette = palette
        self.entities = entities
        self.block_entities = block_entities

    def __eq__(self, other):
        return (
            isinstance(other, ConstructionSection)
            and self.sx == other.sx
            and self.sy == other.sy
            and self.sz == other.sz
            and self.shape == other.shape
            and numpy.equal(self.blocks, other.blocks).all()
            and self.entities == other.entities
            and self.block_entities == other.block_entities
        )

    @property
    def location(self) -> BlockCoordinates:
        return self.sx, self.sy, self.sz



import numpy
from typing import List

from amulet_nbt import CompoundTag

from amulet.api.selection import SelectionBox
from .data_types import BlockDataArrayType, BlockArrayType


class SchematicChunk:
    def __init__(
        self,
        selection: SelectionBox,
        blocks: BlockArrayType,
        data: BlockDataArrayType,
        block_entities: List[CompoundTag],
        entities: List[CompoundTag],
    ):
        self.selection = selection
        self.blocks = blocks
        self.data = data
        self.block_entities = block_entities
        self.entities = entities

    @property
    def blocks(self) -> BlockArrayType:
        return self._blocks

    @blocks.setter
    def blocks(self, blocks: BlockArrayType):
        if not (
            isinstance(blocks, numpy.ndarray)
            and blocks.shape == self.selection.shape
            and blocks.dtype == numpy.uint16
        ):
            raise TypeError(
                "SchematicChunk.blocks must be a uint16 numpy array with shape that matches selection"
            )
        self._blocks = blocks

    @property
    def data(self) -> BlockDataArrayType:
        return self._data

    @data.setter
    def data(self, data: BlockDataArrayType):
        if not (
            isinstance(data, numpy.ndarray)
            and data.shape == self.selection.shape
            and data.dtype == numpy.uint8
        ):
            raise TypeError(
                "SchematicChunk.data must be a uint8 numpy array with shape that matches selection"
            )
        self._data = data



from __future__ import annotations

from .base_level import BaseLevel
from amulet.api import wrapper as api_wrapper


class Structure(BaseLevel):
    """
    Class that handles editing of any structure format via an separate and flexible data format
    """

    def __init__(
        self,
        directory: str,
        structure_wrapper: api_wrapper.StructureFormatWrapper,
    ):
        super().__init__(directory, structure_wrapper)

    @property
    def level_wrapper(self) -> api_wrapper.StructureFormatWrapper:
        """A class to access data directly from the level."""
        return self._level_wrapper



from abc import abstractmethod
from typing import Generator

from amulet.utils.generator import generator_unpacker

from .base_history import BaseHistory


class HistoryManager(BaseHistory):
    """The base class for all active history manager objects.
    The HistoryManager tracks which objects have changed but is not aware of how they have changed."""

    def create_undo_point(self) -> bool:
        """
        Find what has changed since the last undo point and optionally create a new undo point.

        :return: Was an undo point created. If there were no changes no snapshot will be created.
        """
        return generator_unpacker(self.create_undo_point_iter())

    @abstractmethod
    def create_undo_point_iter(self) -> Generator[float, None, bool]:
        """
        Find what has changed since the last undo point and optionally create a new undo point.

        :return: Was an undo point created. If there were no changes no snapshot will be created.
        """
        raise NotImplementedError

    @abstractmethod
    def restore_last_undo_point(self):
        """
        Restore the world to the state it was when self.create_undo_point was called.

        If an operation errors there may be modifications made that did not get tracked.

        This will revert those changes.
        """
        raise NotImplementedError

    @abstractmethod
    def purge(self):
        """Unload all cached data. Effectively returns the class to its starting state."""
        raise NotImplementedError

    @property
    @abstractmethod
    def undo_count(self) -> int:
        """The number of times the :meth:`undo` method can be run."""
        raise NotImplementedError

    @property
    @abstractmethod
    def redo_count(self) -> int:
        """The number of times the :meth:`redo` method can be run."""
        raise NotImplementedError

    @property
    @abstractmethod
    def unsaved_changes(self) -> int:
        """The number of changes that have been made since the last save"""
        raise NotImplementedError
